@@ -11,7 +11,12 @@
 #include "vtkF3DRendererWithColoring.h"
 
 #include <vtkCallbackCommand.h>
+#include <vtkCellPicker.h>
+#include <vtkMath.h>
+#include <vtkMatrix3x3.h>
 #include <vtkNew.h>
+#include <vtkPicker.h>
+#include <vtkPointPicker.h>
 #include <vtkRenderWindow.h>
 #include <vtkRenderWindowInteractor.h>
 #include <vtkRendererCollection.h>
@@ -19,8 +24,12 @@
 #include <vtkVersion.h>
 #include <vtksys/SystemTools.hxx>
 
+#include <chrono>
+#include <cmath>
 #include <map>
 
+#include "camera.h"
+
 namespace f3d::detail
 {
 class interactor_impl::internals
@@ -49,6 +58,16 @@ public:
     dropFilesCallback->SetCallback(OnDropFiles);
     this->Style->AddObserver(vtkF3DInteractorStyle::DropFilesEvent, dropFilesCallback);
 
+    vtkNew<vtkCallbackCommand> middleButtonPressCallback;
+    middleButtonPressCallback->SetClientData(this);
+    middleButtonPressCallback->SetCallback(OnMiddleButtonPress);
+    this->Style->AddObserver(vtkCommand::MiddleButtonPressEvent, middleButtonPressCallback);
+
+    vtkNew<vtkCallbackCommand> middleButtonReleaseCallback;
+    middleButtonReleaseCallback->SetClientData(this);
+    middleButtonReleaseCallback->SetCallback(OnMiddleButtonRelease);
+    this->Style->AddObserver(vtkCommand::MiddleButtonReleaseEvent, middleButtonReleaseCallback);
+
 // Clear needs https://gitlab.kitware.com/vtk/vtk/-/merge_requests/9229
 #if VTK_VERSION_NUMBER >= VTK_VERSION_CHECK(9, 1, 20220601)
     this->Recorder = vtkSmartPointer<vtkF3DInteractorEventRecorder>::New();
@@ -56,8 +75,76 @@ public:
 #endif
   }
 
+  //----------------------------------------------------------------------------
+  // Method defined to normalize the Z axis so all models are treated temporarily
+  // as Z-up axis models.
+  void ToEnvironmentSpace(vtkMatrix3x3* transform)
+  {
+    vtkRenderer* renderer =
+      this->VTKInteractor->GetRenderWindow()->GetRenderers()->GetFirstRenderer();
+    const double* up = renderer->GetEnvironmentUp();
+    const double* right = renderer->GetEnvironmentRight();
+    double fwd[3];
+    vtkMath::Cross(right, up, fwd);
+    const double m[9] = {
+      right[0], right[1], right[2], //
+      fwd[0], fwd[1], fwd[2],       //
+      up[0], up[1], up[2],          //
+    };
+    transform->DeepCopy(m);
+  }
+
+  //----------------------------------------------------------------------------
+  // Set the view orbit position on the viewport.
+  enum class ViewType
+  {
+    VT_FRONT,
+    VT_RIGHT,
+    VT_TOP,
+    VT_ISOMETRIC
+  };
+  static void SetViewOrbit(ViewType view, internals* self)
+  {
+    vtkNew<vtkMatrix3x3> transform;
+    self->ToEnvironmentSpace(transform);
+    camera& cam = self->Window.getCamera();
+    vector3_t up = { 0, 0, 1 };
+    point3_t foc = cam.getFocalPoint();
+    point3_t axis, newPos;
+
+    switch (view)
+    {
+      case ViewType::VT_FRONT:
+        axis = { 0, +1, 0 };
+        break;
+      case ViewType::VT_RIGHT:
+        axis = { +1, 0, 0 };
+        break;
+      case ViewType::VT_TOP:
+        axis = { 0, 0, +1 };
+        up = { 0, -1, 0 };
+        break;
+      case ViewType::VT_ISOMETRIC:
+        axis = { -1, +1, +1 };
+        break;
+    }
+
+    transform->MultiplyPoint(up.data(), up.data());
+    transform->MultiplyPoint(axis.data(), axis.data());
+
+    newPos[0] = foc[0] + axis[0];
+    newPos[1] = foc[1] + axis[1];
+    newPos[2] = foc[2] + axis[2];
+
+    /* set camera coordinates back */
+    cam.setPosition(newPos);
+    cam.setViewUp(up);
+    cam.resetToBounds(0.9);
+  }
+
   static void OnKeyPress(vtkObject*, unsigned long, void* clientData, void*)
   {
+
     internals* self = static_cast<internals*>(clientData);
     vtkRenderWindowInteractor* rwi = self->Style->GetInteractor();
     int keyCode = std::toupper(rwi->GetKeyCode());
@@ -80,12 +167,19 @@ public:
     bool checkColoring = false;
     bool render = false;
 
+    // Available keycodes: None
     switch (keyCode)
     {
+      case 'W':
+        self->AnimationManager->CycleAnimation();
+        self->Options.set("scene.animation.index", self->AnimationManager->GetAnimationIndex());
+        ren->SetAnimationnameInfo(self->AnimationManager->GetAnimationName());
+        render = true;
+        break;
       case 'C':
         if (renWithColor)
         {
-          renWithColor->CycleScalars(vtkF3DRendererWithColoring::F3D_FIELD_CYCLE);
+          renWithColor->CycleScalars(vtkF3DRendererWithColoring::CycleType::FIELD);
           self->Window.PrintColoringDescription(log::VerboseLevel::DEBUG);
           checkColoring = true;
           render = true;
@@ -94,7 +188,7 @@ public:
       case 'S':
         if (renWithColor)
         {
-          renWithColor->CycleScalars(vtkF3DRendererWithColoring::F3D_ARRAY_CYCLE);
+          renWithColor->CycleScalars(vtkF3DRendererWithColoring::CycleType::ARRAY_INDEX);
           self->Window.PrintColoringDescription(log::VerboseLevel::DEBUG);
           checkColoring = true;
           render = true;
@@ -103,7 +197,7 @@ public:
       case 'Y':
         if (renWithColor)
         {
-          renWithColor->CycleScalars(vtkF3DRendererWithColoring::F3D_COMPONENT_CYCLE);
+          renWithColor->CycleScalars(vtkF3DRendererWithColoring::CycleType::COMPONENT);
           self->Window.PrintColoringDescription(log::VerboseLevel::DEBUG);
           checkColoring = true;
           render = true;
@@ -113,17 +207,16 @@ public:
         self->Options.toggle("ui.bar");
         render = true;
         break;
-      case 'p':
       case 'P':
-        self->Options.toggle("render.effect.depth-peeling");
+        self->Options.toggle("render.effect.translucency-support");
         render = true;
         break;
       case 'Q':
-        self->Options.toggle("render.effect.ssao");
+        self->Options.toggle("render.effect.ambient-occlusion");
         render = true;
         break;
       case 'A':
-        self->Options.toggle("render.effect.fxaa");
+        self->Options.toggle("render.effect.anti-aliasing");
         render = true;
         break;
       case 'T':
@@ -139,7 +232,7 @@ public:
         render = true;
         break;
       case 'G':
-        self->Options.toggle("render.grid");
+        self->Options.toggle("render.grid.enable");
         render = true;
         break;
       case 'N':
@@ -184,6 +277,38 @@ public:
         self->Options.toggle("interactor.trackball");
         render = true;
         break;
+      case 'F':
+        self->Options.toggle("render.hdri.ambient");
+        render = true;
+        break;
+      case 'J':
+        self->Options.toggle("render.background.skybox");
+        render = true;
+        break;
+      case 'L':
+      {
+        const double intensity = self->Options.getAsDouble("render.light.intensity");
+        const bool down = rwi->GetShiftKey();
+
+        /* `ref < x` is equivalent to:
+         * - `intensity <= x` when going down
+         * - `intensity < x` when going up */
+        const double ref = down ? intensity - 1e-6 : intensity;
+        // clang-format off
+        /* offset in percentage points */
+        const int offsetPp = ref < .5 ?  1
+                           : ref <  1 ?  2
+                           : ref <  5 ?  5
+                           : ref < 10 ? 10
+                           :            25;
+        // clang-format on
+        /* new intensity in percents */
+        const int newIntensityPct = std::lround(intensity * 100) + (down ? -offsetPp : +offsetPp);
+
+        self->Options.set("render.light.intensity", std::max(newIntensityPct, 0) / 100.0);
+        render = true;
+        break;
+      }
       case 'H':
         self->Options.toggle("ui.cheatsheet");
         render = true;
@@ -192,29 +317,28 @@ public:
         self->Window.PrintColoringDescription(log::VerboseLevel::INFO);
         self->Window.PrintSceneDescription(log::VerboseLevel::INFO);
         break;
+      case '1':
+        self->SetViewOrbit(ViewType::VT_FRONT, self);
+        render = true;
+        break;
+      case '3':
+        self->SetViewOrbit(ViewType::VT_RIGHT, self);
+        render = true;
+        break;
+      case '5':
+        self->Options.toggle("scene.camera.orthographic");
+        render = true;
+        break;
+      case '7':
+        self->SetViewOrbit(ViewType::VT_TOP, self);
+        render = true;
+        break;
+      case '9':
+        self->SetViewOrbit(ViewType::VT_ISOMETRIC, self);
+        render = true;
+        break;
       default:
-        if (keySym == "Left")
-        {
-          self->AnimationManager.StopAnimation();
-          loader::LoadFileEnum load = loader::LoadFileEnum::LOAD_PREVIOUS;
-          self->Loader.loadFile(load);
-          render = true;
-        }
-        else if (keySym == "Right")
-        {
-          self->AnimationManager.StopAnimation();
-          loader::LoadFileEnum load = loader::LoadFileEnum::LOAD_NEXT;
-          self->Loader.loadFile(load);
-          render = true;
-        }
-        else if (keySym == "Up")
-        {
-          self->AnimationManager.StopAnimation();
-          loader::LoadFileEnum load = loader::LoadFileEnum::LOAD_CURRENT;
-          self->Loader.loadFile(load);
-          render = true;
-        }
-        else if (keySym == F3D_EXIT_HOTKEY_SYM)
+        if (keySym == F3D_EXIT_HOTKEY_SYM)
         {
           self->StopInteractor();
         }
@@ -225,7 +349,8 @@ public:
         }
         else if (keySym == "Space")
         {
-          self->AnimationManager.ToggleAnimation();
+          assert(self->AnimationManager);
+          self->AnimationManager->ToggleAnimation();
         }
         break;
     }
@@ -248,9 +373,10 @@ public:
     internals* self = static_cast<internals*>(clientData);
     vtkStringArray* filesArr = static_cast<vtkStringArray*>(callData);
     std::vector<std::string> filesVec;
+    filesVec.resize(filesArr->GetNumberOfTuples());
     for (int i = 0; i < filesArr->GetNumberOfTuples(); i++)
     {
-      filesVec.push_back(filesArr->GetValue(i));
+      filesVec[i] = filesArr->GetValue(i);
     }
 
     if (self->DropFilesUserCallBack(filesVec))
@@ -258,22 +384,117 @@ public:
       return;
     }
 
-    // No user defined behavior, use standard behavior
-    self->AnimationManager.StopAnimation();
-    for (std::string file : filesVec)
+    // No user defined behavior, load the first file
+    if (!filesVec.empty())
     {
-      self->Loader.addFile(file);
+      assert(self->AnimationManager);
+      self->AnimationManager->StopAnimation();
+      if (self->Loader.hasSceneReader(filesVec[0]))
+      {
+        self->Loader.loadScene(filesVec[0]);
+      }
+      else if (self->Loader.hasGeometryReader(filesVec[0]))
+      {
+        self->Loader.loadGeometry(filesVec[0], true);
+      }
+      self->Window.render();
     }
-    self->Loader.loadFile(loader::LoadFileEnum::LOAD_LAST);
-    self->Window.render();
   }
 
-  std::function<bool(int, std::string)> KeyPressUserCallBack = [](int, std::string)
-  { return false; };
-  std::function<bool(std::vector<std::string>)> DropFilesUserCallBack = [](std::vector<std::string>)
-  { return false; };
+  static void OnMiddleButtonPress(vtkObject*, unsigned long, void* clientData, void*)
+  {
+    internals* self = static_cast<internals*>(clientData);
 
-  void StartInteractor() { this->VTKInteractor->Start(); }
+    self->VTKInteractor->GetEventPosition(self->MiddleButtonDownPosition);
+
+    self->Style->OnMiddleButtonDown();
+  }
+
+  static void OnMiddleButtonRelease(vtkObject*, unsigned long, void* clientData, void*)
+  {
+    internals* self = static_cast<internals*>(clientData);
+
+    const int* middleButtonUpPosition = self->VTKInteractor->GetEventPosition();
+
+    const int xDelta = middleButtonUpPosition[0] - self->MiddleButtonDownPosition[0];
+    const int yDelta = middleButtonUpPosition[1] - self->MiddleButtonDownPosition[1];
+    const int sqPosDelta = xDelta * xDelta + yDelta * yDelta;
+    if (sqPosDelta < self->DragDistanceTol * self->DragDistanceTol)
+    {
+      const int x = self->MiddleButtonDownPosition[0];
+      const int y = self->MiddleButtonDownPosition[1];
+      vtkRenderer* renderer =
+        self->VTKInteractor->GetRenderWindow()->GetRenderers()->GetFirstRenderer();
+
+      bool pickSuccessful = false;
+      double picked[3];
+      if (self->CellPicker->Pick(x, y, 0, renderer))
+      {
+        self->CellPicker->GetPickPosition(picked);
+        pickSuccessful = true;
+      }
+      else if (self->PointPicker->Pick(x, y, 0, renderer))
+      {
+        self->PointPicker->GetPickPosition(picked);
+        pickSuccessful = true;
+      }
+
+      if (pickSuccessful)
+      {
+        /*     pos.--------------------.foc
+         *       /|                   /
+         *      / |                  /
+         *     .--.-----------------.picked
+         * pos1    pos2
+         */
+        const camera_state_t state = self->Window.getCamera().getState();
+
+        double focV[3];
+        vtkMath::Subtract(picked, state.foc.data(), focV); /* foc -> picked */
+
+        double posV[3];
+        vtkMath::Subtract(picked, state.foc.data(), posV); /* pos -> pos1, parallel to focV */
+        if (!self->Style->GetInteractor()->GetShiftKey())
+        {
+          double v[3];
+          vtkMath::Subtract(state.foc.data(), state.pos.data(), v); /* pos -> foc */
+          vtkMath::ProjectVector(focV, v, v);                       /* pos2 -> pos1 */
+          vtkMath::Subtract(posV, v, posV); /* pos -> pos2, keeps on camera plane */
+        }
+
+        const auto interpolateCameraState = [&state, &focV, &posV](double ratio) -> camera_state_t
+        {
+          return { //
+            {
+              state.pos[0] + posV[0] * ratio,
+              state.pos[1] + posV[1] * ratio,
+              state.pos[2] + posV[2] * ratio,
+            },
+            {
+              state.foc[0] + focV[0] * ratio,
+              state.foc[1] + focV[1] * ratio,
+              state.foc[2] + focV[2] * ratio,
+            },
+            state.up, state.angle
+          };
+        };
+
+        self->AnimateCameraTransition(interpolateCameraState);
+      }
+    }
+
+    self->Style->OnMiddleButtonUp();
+  }
+
+  std::function<bool(int, const std::string&)> KeyPressUserCallBack = [](int, const std::string&)
+  { return false; };
+  std::function<bool(const std::vector<std::string>&)> DropFilesUserCallBack =
+    [](const std::vector<std::string>&) { return false; };
+
+  void StartInteractor()
+  {
+    this->VTKInteractor->Start();
+  }
 
   void StopInteractor()
   {
@@ -281,25 +502,66 @@ public:
     this->VTKInteractor->ExitCallback();
   }
 
+  /**
+   * Run a camera transition animation based on a camera state interpolation function.
+   * The provided function will be called with an interpolation parameter
+   * varying from `0.` for the initial state to `1.` for the final state;
+   * it shall return an appropriate linearly interpolated `camera_state_t` for any value in between.
+   */
+  template<class CameraStateInterpolator>
+  void AnimateCameraTransition(CameraStateInterpolator interpolateCameraState)
+  {
+    window& win = this->Window;
+    camera& cam = win.getCamera();
+    const int duration = this->TransitionDuration;
+
+    if (duration > 0)
+    {
+      // TODO implement a way to not queue key presses while the animation is running
+
+      const auto start = std::chrono::high_resolution_clock::now();
+      const auto end = start + std::chrono::milliseconds(duration);
+      auto now = start;
+      while (now < end)
+      {
+        const double timeDelta =
+          std::chrono::duration_cast<std::chrono::milliseconds>(now - start).count();
+        const double ratio = (1 - std::cos(vtkMath::Pi() * (timeDelta / duration))) / 2;
+        cam.setState(interpolateCameraState(ratio));
+        this->Window.render();
+        now = std::chrono::high_resolution_clock::now();
+      }
+    }
+
+    cam.setState(interpolateCameraState(1.)); // ensure final update
+    this->Window.render();
+  }
+
   options& Options;
   window_impl& Window;
   loader_impl& Loader;
-  animationManager AnimationManager;
+  animationManager* AnimationManager;
 
   vtkNew<vtkRenderWindowInteractor> VTKInteractor;
   vtkNew<vtkF3DInteractorStyle> Style;
   vtkSmartPointer<vtkF3DInteractorEventRecorder> Recorder;
-  int WindowSize[2] = { -1, -1 };
-  int WindowPos[2] = { 0, 0 };
   std::map<unsigned long, std::pair<int, std::function<void()> > > TimerCallBacks;
+
+  vtkNew<vtkCellPicker> CellPicker;
+  vtkNew<vtkPointPicker> PointPicker;
+
+  int MiddleButtonDownPosition[2] = { 0, 0 };
+
+  int DragDistanceTol = 3;      /* px */
+  int TransitionDuration = 100; /* ms */
 };
 
 //----------------------------------------------------------------------------
 interactor_impl::interactor_impl(options& options, window_impl& window, loader_impl& loader)
   : Internals(std::make_unique<interactor_impl::internals>(options, window, loader))
 {
-  // Loader need the interactor
-  this->Internals->Loader.setInteractor(this);
+  // Loader need the interactor, loader will set the AnimationManager on the interactor
+  this->Internals->Loader.SetInteractor(this);
 }
 
 //----------------------------------------------------------------------------
@@ -353,25 +615,29 @@ unsigned long interactor_impl::createTimerCallBack(double time, std::function<vo
 //----------------------------------------------------------------------------
 void interactor_impl::toggleAnimation()
 {
-  this->Internals->AnimationManager.ToggleAnimation();
+  assert(this->Internals->AnimationManager);
+  this->Internals->AnimationManager->ToggleAnimation();
 }
 
 //----------------------------------------------------------------------------
 void interactor_impl::startAnimation()
 {
-  this->Internals->AnimationManager.StartAnimation();
+  assert(this->Internals->AnimationManager);
+  this->Internals->AnimationManager->StartAnimation();
 }
 
 //----------------------------------------------------------------------------
 void interactor_impl::stopAnimation()
 {
-  this->Internals->AnimationManager.StopAnimation();
+  assert(this->Internals->AnimationManager);
+  this->Internals->AnimationManager->StopAnimation();
 }
 
 //----------------------------------------------------------------------------
 bool interactor_impl::isPlayingAnimation()
 {
-  return this->Internals->AnimationManager.IsPlaying();
+  assert(this->Internals->AnimationManager);
+  return this->Internals->AnimationManager->IsPlaying();
 }
 
 //----------------------------------------------------------------------------
@@ -427,11 +693,28 @@ bool interactor_impl::recordInteraction(const std::string& file)
 {
   if (file.empty())
   {
-    // TODO improve VTK to check file opening
     log::error("No interaction record file provided");
     return false;
   }
 
+  std::string cleanFile = vtksys::SystemTools::CollapseFullPath(file);
+
+  std::string parentDirectory = vtksys::SystemTools::GetParentDirectory(cleanFile);
+
+  // Check if the parent directory exists
+  if (!vtksys::SystemTools::FileExists(parentDirectory))
+  {
+    log::error("Interaction record directory does not exist ", parentDirectory);
+    return false;
+  }
+
+  // Check if we can write to the directory
+  if (!vtksys::SystemTools::TestFileAccess(parentDirectory, vtksys::TEST_FILE_WRITE))
+  {
+    log::error("Don't have write permissions for ", parentDirectory);
+    return false;
+  }
+
 // Clear needs https://gitlab.kitware.com/vtk/vtk/-/merge_requests/9229
 #if VTK_VERSION_NUMBER >= VTK_VERSION_CHECK(9, 1, 20220601)
   // Make sure the recorder is off and streams are cleared
@@ -443,7 +726,6 @@ bool interactor_impl::recordInteraction(const std::string& file)
   this->Internals->Recorder->SetInteractor(this->Internals->VTKInteractor);
 #endif
 
-  std::string cleanFile = vtksys::SystemTools::CollapseFullPath(file);
   this->Internals->Recorder->SetFileName(cleanFile.c_str());
   this->Internals->Recorder->On();
   this->Internals->Recorder->Record();
@@ -464,6 +746,12 @@ void interactor_impl::stop()
   this->Internals->StopInteractor();
 }
 
+//----------------------------------------------------------------------------
+void interactor_impl::SetAnimationManager(animationManager* manager)
+{
+  this->Internals->AnimationManager = manager;
+}
+
 //----------------------------------------------------------------------------
 void interactor_impl::SetInteractorOn(vtkInteractorObserver* observer)
 {
@@ -471,9 +759,8 @@ void interactor_impl::SetInteractorOn(vtkInteractorObserver* observer)
 }
 
 //----------------------------------------------------------------------------
-void interactor_impl::InitializeAnimation(vtkImporter* importer)
+void interactor_impl::UpdateRendererAfterInteraction()
 {
-  this->Internals->AnimationManager.Initialize(
-    &this->Internals->Options, this, &this->Internals->Window, importer);
+  this->Internals->Style->UpdateRendererAfterInteraction();
 }
 }
