@@ -13,8 +13,9 @@ int TestSDKDynamicBackgroundColor(int argc, char* argv[])
   f3d::options& opt = eng.getOptions();
   win.setSize(300, 300);
   opt.set("ui.filename", true);
+  opt.set("ui.filename-info", "(1/1) cow.vtp");
 
-  load.addFile(std::string(argv[1]) + "/data/cow.vtp").loadFile();
+  load.loadGeometry(std::string(argv[1]) + "/data/cow.vtp");
 
   win.render();
 
